"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const logger_1 = require("@pnpm/logger");
const lock_1 = __importDefault(require("./lock"));
const http = require("http");
function default_1(store, opts) {
    const rawManifestPromises = {};
    const filesPromises = {};
    // eslint-disable-next-line @typescript-eslint/no-invalid-void-type
    const lock = lock_1.default();
    const server = http.createServer(async (req, res) => {
        var _a;
        if (req.method !== 'POST') {
            res.statusCode = 405; // Method Not Allowed
            const responseError = { error: `Only POST is allowed, received ${(_a = req.method) !== null && _a !== void 0 ? _a : 'unknown'}` };
            res.setHeader('Allow', 'POST');
            res.end(JSON.stringify(responseError));
            return;
        }
        const bodyPromise = new Promise((resolve, reject) => {
            let body = ''; // eslint-disable-line
            req.on('data', (data) => {
                body += data; // eslint-disable-line
            });
            req.on('end', async () => {
                try {
                    if (body.length > 0) {
                        body = JSON.parse(body);
                    }
                    else {
                        body = {};
                    }
                    resolve(body);
                }
                catch (e) {
                    reject(e);
                }
            });
        });
        try {
            let body;
            switch (req.url) {
                case '/requestPackage': {
                    try {
                        body = await bodyPromise;
                        const pkgResponse = await store.requestPackage(body.wantedDependency, body.options);
                        if (pkgResponse['bundledManifest']) { // eslint-disable-line
                            rawManifestPromises[body.msgId] = pkgResponse['bundledManifest']; // eslint-disable-line
                            pkgResponse.body['fetchingBundledManifestInProgress'] = true; // eslint-disable-line
                        }
                        if (pkgResponse['files']) { // eslint-disable-line
                            filesPromises[body.msgId] = pkgResponse['files']; // eslint-disable-line
                        }
                        res.end(JSON.stringify(pkgResponse.body));
                    }
                    catch (err) {
                        res.end(JSON.stringify({
                            error: {
                                message: err.message,
                                ...JSON.parse(JSON.stringify(err)),
                            },
                        }));
                    }
                    break;
                }
                case '/fetchPackage': {
                    try {
                        body = await bodyPromise;
                        const pkgResponse = store.fetchPackage(body.options);
                        if (pkgResponse['bundledManifest']) { // eslint-disable-line
                            rawManifestPromises[body.msgId] = pkgResponse['bundledManifest']; // eslint-disable-line
                        }
                        if (pkgResponse['files']) { // eslint-disable-line
                            filesPromises[body.msgId] = pkgResponse['files']; // eslint-disable-line
                        }
                        res.end(JSON.stringify({ filesIndexFile: pkgResponse.filesIndexFile }));
                    }
                    catch (err) {
                        res.end(JSON.stringify({
                            error: {
                                message: err.message,
                                ...JSON.parse(JSON.stringify(err)),
                            },
                        }));
                    }
                    break;
                }
                case '/packageFilesResponse': {
                    body = await bodyPromise;
                    const filesResponse = await filesPromises[body.msgId]();
                    delete filesPromises[body.msgId];
                    res.end(JSON.stringify(filesResponse));
                    break;
                }
                case '/rawManifestResponse': {
                    body = await bodyPromise;
                    const manifestResponse = await rawManifestPromises[body.msgId]();
                    delete rawManifestPromises[body.msgId];
                    res.end(JSON.stringify(manifestResponse));
                    break;
                }
                case '/prune':
                    // Disable store pruning when a server is running
                    res.statusCode = 403;
                    res.end();
                    break;
                case '/importPackage': {
                    const importPackageBody = (await bodyPromise); // eslint-disable-line @typescript-eslint/no-explicit-any
                    await store.importPackage(importPackageBody.to, importPackageBody.opts);
                    res.end(JSON.stringify('OK'));
                    break;
                }
                case '/upload': {
                    // Do not return an error status code, just ignore the upload request entirely
                    if (opts.ignoreUploadRequests) {
                        res.statusCode = 403;
                        res.end();
                        break;
                    }
                    const uploadBody = (await bodyPromise); // eslint-disable-line @typescript-eslint/no-explicit-any
                    await lock(uploadBody.builtPkgLocation, () => store.upload(uploadBody.builtPkgLocation, uploadBody.opts));
                    res.end(JSON.stringify('OK'));
                    break;
                }
                case '/stop':
                    if (opts.ignoreStopRequests) {
                        res.statusCode = 403;
                        res.end();
                        break;
                    }
                    logger_1.globalInfo('Got request to stop the server');
                    await close();
                    res.end(JSON.stringify('OK'));
                    logger_1.globalInfo('Server stopped');
                    break;
                default: {
                    res.statusCode = 404;
                    const error = { error: `${req.url} does not match any route` };
                    res.end(JSON.stringify(error));
                }
            }
        }
        catch (e) {
            res.statusCode = 503;
            const jsonErr = JSON.parse(JSON.stringify(e));
            jsonErr.message = e.message;
            res.end(JSON.stringify(jsonErr));
        }
    });
    let listener;
    if (opts.path) {
        listener = server.listen(opts.path);
    }
    else {
        listener = server.listen(opts.port, opts.hostname);
    }
    return { close };
    function close() {
        listener.close();
        return store.close();
    }
}
exports.default = default_1;
//# sourceMappingURL=createServer.js.map